\\ Copyright 2014, 2015 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.



gcdext(x,y)=bezout(x,y);
vecsum(v)=sum(i=1,#v,v[i]);

\\----------------------------------------------------------------------------

\\ p is a polynomial
\\ If a coefficient of p is a t_COMPLEX or t_QUAD then return it.
pol_partial_fractions__find_factor(p) =
{
  for(i=0,poldegree(p),
     my(c = polcoeff(p,i));
     if(c && (type(c) == "t_COMPLEX" || type(c) == "t_QUAD"),
        return(if(c==0,c+1,c))));
  0;
}

pol_partial_fractions(f, dens=0) =
{
  if(f==0, return([]));
  my(n = numerator(f));

  \\ dens by factor(), and possible complex or quad
  if(dens == 0 || type(dens) == "t_COMPLEX" || type(dens) == "t_QUAD",
     dens*f;  \\ check dens complex or quad is compatible with f
     my(d = denominator(f));
     if(! pol_partial_fractions__find_factor(d),
        \\ no quad in denominator
        if(dens,
           d *= dens, \\ given dens
           my(q = pol_partial_fractions__find_factor(n));
           if(q, d*=q)));   \\ otherwise quad from numerator
     dens = factor(d));

  \\ if dens is a factor() matrix then take its factors column
  if(type(dens)=="t_MAT", dens=dens[,1]);
  print("dens="dens);

  \\ dens type check
  for(i=1,#dens,
     my(p = dens[i]);
     if(type(p) != "t_POL", error("dens entry not a polynomial"));
     if(poldegree(p) < 1, error("dens entry is a constant")));

  my(ret:list = List([]),
     d = denominator(f));     \\ n/d reduced progressively

  \\ quotient n/d if numerator has higher degree than denominator
  my(v);
  if(poldegree(n) >= poldegree(d),
     v = divrem(n,d);
     listput(ret,v[1]); \\ quotient
     n = v[2];          \\ remainder
     print("num higher degree q=",q," remaining num=",n);
    );

  for(i=1,#dens,
     my(p = dens[i],
        e = valuation(d, p));
     if(e<1, next());  \\ p is not a factor of denominator d
     my(x = p^e,
        y = d/x);      \\ x*y = d
     print("i=",i," p="p"  e="e"  x="x"  y="y);
     print("  n="n"  d="d);

     \\ find a*y + b*x = g
     v = gcdext(y, x);
     my(a=v[1],b=v[2],g=v[3]);
     \\ print("  a="a" b="b"  g="g);
     \\ if(a*y+b*x != g, error("oops gcd"));

     \\ If g!=1 then x and y have a common factor due to p a power like
     \\ (x+1)^2 but denominator(f) a power not a multiple of that like
     \\ (x+1)^3.  gcd((x+1)^2, (x+1)^3) == x+1.  Cannot break out this p.
     \\ (Since n and d have no common factor cannot multiply n/g to give n.)
     \\
     if(v[3]!=1,next());

     \\ multiply by n to give a*y + b*x = n
     a *= n;
     b *= n;

     \\ reduce a to poldegree(a) < poldegree(x) by mod x and corresponding
     \\ adjustment to b to maintain a*y + b*x = n
     v = divrem(a,x);
     my(q=v[1]);
     a = v[2];
     b += q*y;
     \\ print("  mul a="a" b="b);
     \\ if(a/x+b/y != n/d, error("oops pfrac sum ",a/x," ",b/y));

     \\ now have a/x + b/y = n/d with suitable poldegree() of a and b
     n = b;
     d = y;

     \\ If e==1 then a/x is the partial fraction.
     \\ If e>1 then break further a[e]/p^e, ..., a[2]/p^2, a[1]/p.
     my(pos = #ret+1);  \\ insert so return ascending powers
     while(a && e>1,
           v = divrem(a,p);
           a=v[1];
           my(r=v[2]);
           if(r, listinsert(ret, r/x, pos));
           x /= p;
           e--);
     print("  pfrac a="a" / x="x);
     if(a, listinsert(ret, a/x, pos)));


  \\ final remaining n/d, if any
  if(n, listput(ret,n/d));
  n = Vec(ret);
  print("final ",n);

  \\ add-back as a check
  my(s = sum(i=1,#n,n[i]));
  if(s != f,
     error("pol_partial_fractions() oops, partial fractions don't sum\nret=",n,"\nsum got   ",s,"\nwant orig ",f));

  n;
}

{
addhelp(pol_partial_fractions,
"v = pol_partial_fractions(f, dens=0)
Return a vector which is polynomial fraction f broken into partial fractions.

\tf = 2*x/(x^2 - 1);
\tv = pol_partial_fractions(f);
gives
\tv == [1/(x-1), 1/(x+1)]
\tvecsum(v) == f

Optional argument dens is a vector of denominators to try to break out as
partial fractions.  The default is factor() on the denominator of f.

See comments at the start of pol-pfrac.gp for more.");
}

\\-----------------------------------------------------------------------------
\\ v is a vector of polynomial fractions.
\\ Merge terms with the same denominator by adding.
\\ The return is a new reduced vector.
\\
pol_partial_fractions_merge(v) =
{
  my(orig_v = v);
  for(i=2,#v,
     my(p=v[i]);
     for(j=1,i-1,
        if(v[j] && poldegree(denominator(p)) == poldegree(denominator(v[j])),
           my(s = p + v[j]);
           if(poldegree(denominator(s)) <= poldegree(denominator(p)),
              v[j] = s;      \\ merge v[i] into earlier entry v[j]
              v[i] = 0;
              next(2)))));   \\ next v[i]

  v = select((elem)->elem,v); \\ non-zero elements
  print("v = ",v);

  \\ add-back as a check
  if(sum(i=1,#v,v[i]) != sum(i=1,#orig_v,orig_v[i]),
     error("oops merged fractions don't sum, new v=",v,
           " sum got ",sum(i=1,#v,v[i])," want ",sum(i=1,#orig_v,orig_v[i])));
  v;
}
{
addhelp(pol_partial_fractions_merge,
"v = pol_partial_fractions_merge(v)
v is a vector of polynomial fractions.
Return a new vector which has fractions with the same denominators summed
together.  For example,

\tpol_partial_fractions_merge([1/(x-1), x/(x-1), 1/(x+1)])
gives
\t[(x+1)/(x-1), 1/(x+1)]

See comments at the start of pol-pfrac.gp for more.");
}

\\-----------------------------------------------------------------------------

\\ gp-2.5.5 pol-255.gp ../test.gp
